#!/usr/bin/env python

from __future__ import absolute_import

import migrate.versioning.api as dbmigrate

_just_a_simulation = "%s: Don't worry: this is just a simulation."


class doctests(object):

    def __init__(self, packages):
        self.packages = packages

    def run(self):
        from distutils.util import convert_path
        import doctest as _doctest

        for package in self.packages:
            for source in findfile(kind='f', names=["*.py"], topdown=False):
                _doctest.testfile(source, False)
    

class zap(object):

    def __init__(self, c_files=False, cpp_files=False):
        self.files = '*~,*.pyc,*.so,*.a'
        self.dirs  = 'build,dist,*.egg,*.egg-info,__pycache__'
        if c_files:   self.files += ',*.c,*.h'
        if cpp_files: self.files += ',*.cpp,*.hpp'
        # convert to list
        self.dirs  = self.__split(self.dirs)
        self.files = self.__split(self.files)
        self.dry_run = False
        self.verbose = False

    def set_dry_run(self, dry_run):
        self.dry_run = dry_run
        if self.dry_run: print(_just_a_simulation % "zap")

    def set_verbose(self, verbose):
        self.verbose = verbose
        if self.verbose:
            print("zap: files=%s" % ( self.files ) )
            print("zap: dirs=%s"  % ( self.dirs ) )

    def run(self):
        if not self.dry_run:
            # remove files
            for base in self.dirs:
                self.__remove(base=base)
                # remove directories
                self.__remove(names=self.files, kind='f')
                # remove __pycache__
                for cache in findfile(names=['__pycache__'], kind='d', topdown=False):
                    self.__remove(base=cache)

    def __remove(self, base='.', names=[], kind=None):
        import os
        for fn in findfile(base=base, names=names, kind=kind, topdown=False):
            if os.path.isfile(fn):
                if self.dry_run:
                    print('rm    ' + fn)
                else:
                    os.remove(fn)
            elif os.path.isdir(fn):
                if self.dry_run:
                    print('rmdir ' + fn)
                else:
                    os.rmdir(fn)

    import re
    __spaces = re.compile( ',[ \t]+' )

    def __split(self, s):
        s = self.__spaces.sub(',', s)
        return map((lambda x: x.strip()), s.split(','))


#TODO: allow operations on remote databases
class psql(object):

    def __init__(self, url=None):
        self.url = url
        self.dry_run = False
        self.verbose = False

    def set_dry_run(self, dry_run):
        self.dry_run = dry_run
        if self.dry_run: print(_just_a_simulation % "psql")

    def set_verbose(self, verbose):
        self.verbose = verbose
        if self.verbose:
            print("psql: url=%s" % ( self.url ) )

    def login(self):
        if not self.dry_run:
            self.__bootstrap()
            self.__call(['/usr/bin/sudo','-k','-S','-u','postgres','/usr/bin/whoami'], self.__passwd)

    def create_user(self):
        if not self.dry_run:
            self.__bootstrap()
            self.__call(['/usr/bin/sudo','-u','postgres',
                         '/usr/bin/createuser','--createdb',self.__username])
            self.__call(['/usr/bin/sudo','-u','postgres',
                         '/usr/bin/psql','-c', 
                         "alter user %s with password '%s'" % (self.__username, self.__password)])

    def create_db(self):
        if not self.dry_run:
            self.__bootstrap()
            self.__call(['/usr/bin/sudo','-u','postgres',
                         '/usr/bin/createdb',self.__database,
                         '--owner',self.__username])

    def drop_user(self):
        if not self.dry_run:
            self.__bootstrap()
            self.__call(['/usr/bin/sudo','-u','postgres',
                         '/usr/bin/dropuser',self.__username])


    def drop_db(self):
        if not self.dry_run:
            self.__bootstrap()
            self.__call(['/usr/bin/sudo','-u','postgres',
                         '/usr/bin/dropdb',self.__database])

    def __bootstrap(self):
        check = dsn(self.url, dry_run=self.dry_run).check()
        self.__username = check['username']
        self.__password = check['password']
        self.__database = check['database']

    def __call(self, args, method=None, verbose=True):
        cmd = ' '.join(args)
        if self.verbose and cmd.find('password') == -1:
            print("psql: cmd=%s" % ( cmd ) )
        if not self.dry_run:
            text = method() if method else None
            import subprocess as sp
            p = sp.Popen(args, shell=False, stdin=sp.PIPE, stdout=sp.PIPE, stderr=sp.PIPE)
            (stdout, stderr) = p.communicate(text)
            if stdout is not None: print(stdout)
            if stderr is not None: print(stderr)

    def __passwd(self):
        import getpass
        return getpass.getpass("Password: ") + '\n'



class migrate(object):

    def __init__(self, url=None, repository='admin/migrate', application=None):
        import sys
        self.application = application if application is not None else sys.argv[0]
        self.repository = repository
        self.url = url
        self.dry_run = False
        self.verbose = False

    def set_dry_run(self, dry_run):
        self.dry_run = dry_run
        if self.dry_run:
            print(_just_a_simulation % "migrate")

    def set_verbose(self, verbose):
        self.verbose = verbose
        if self.verbose:
            print("migrate: application=%s repository=%s url=%s" % (
                self.application, self.repository, self.url ) )

    def create_model(self):
        if not self.dry_run:
            self.__bootstrap()
            result = dbmigrate.create_model(self.url, self.repository)
            v = '%03d' % (self.__model_version() + 1)
            path = self.repository + '/versions/' + v + '_reverse_engineered.py'
            if self.verbose: print("migrate: create_model: path=%s" % ( path ) )
            f = open(path, 'w')
            f.write(result)
            f.close()

    def upgrade(self, changeset=None):
        """
        Applies changesets to a database, ie: upgrades the database.
        If the changeset is not informed, applies all changesets to the database.
        WARNING: this is a POTENTIALLY DANGEROUS operation which MAY DESTROY YOUR DATA.

        >>> # WARNING: SQLite in-memory database seems to not work properly
        >>> #          So, avoid these url: "sqlite://" and "sqlite:///:memory:"
        >>> url = 'sqlite:///build/doctest_upgrade.db'
        >>> import distcontrib.api as api
        >>> migrate = api.migrate(url, 'tests/guinea-pig/admin/migrate')
        >>> migrate.set_verbose(True)
        migrate: application=setup.py repository=tests/guinea-pig/admin/migrate url=sqlite:///build/doctest_upgrade.db
        >>> migrate.upgrade()
        migrate: upgrade: changeset=2
        >>> migrate.status()
        Model version:    2
        Database version: 2
        (2, 2)

        """
        if not self.dry_run:
            self.__bootstrap()
            if changeset is None:
                 changeset = self.__model_version()
            if self.verbose: print("migrate: upgrade: changeset=%s" % ( changeset ))
            dbmigrate.upgrade(self.url, self.repository, version=changeset)

    def downgrade(self, changeset=None):
        """
        Reverts changesets from a database, ie: downgrades the database.
        If the changeset is not informed, reverts all changesets from the database.
        WARNING: this is a POTENTIALLY DANGEROUS operation which MAY DESTROY YOUR DATA.

        >>> # WARNING: SQLite in-memory database seems to not work properly
        >>> #          So, avoid these url: "sqlite://" and "sqlite:///:memory:"
        >>> url = 'sqlite:///build/doctest_downgrade.db'
        >>> import distcontrib.api as api
        >>> migrate = api.migrate(url, 'tests/guinea-pig/admin/migrate')
        >>> migrate.set_verbose(True)
        migrate: application=setup.py repository=tests/guinea-pig/admin/migrate url=sqlite:///build/doctest_downgrade.db
        >>> migrate.upgrade()
        migrate: upgrade: changeset=2
        >>> migrate.downgrade()
        migrate: downgrade: changeset=1
        >>> migrate.status()
        Model version:    2
        Database version: 1
        (2, 1)

        """
        if not self.dry_run:
            self.__bootstrap()
            if changeset is None:
                 changeset = max(0, self.__database_version() - 1)
            if self.verbose: print("migrate: downgrade: changeset=%s" % ( changeset ))
            dbmigrate.downgrade(self.url, self.repository, changeset)

    def sanity(self):
        """
        Runs sanity checks, applying all change sets to a database and then reverting all changesets.
        This is potentially DANGEROUS test which will certainly DESTROY ALL YOUR DATA.
        So, run this test against a test database.

        >>> # WARNING: SQLite in-memory database seems to not work properly
        >>> #          So, avoid these url: "sqlite://" and "sqlite:///:memory:"
        >>> url = 'sqlite:///build/doctest_sanitytest.db'
        >>> import distcontrib.api as api
        >>> migrate = api.migrate(url, 'tests/guinea-pig/admin/migrate')
        >>> migrate.set_verbose(True)
        migrate: application=setup.py repository=tests/guinea-pig/admin/migrate url=sqlite:///build/doctest_sanitytest.db
        >>> migrate.sanity()
        migrate: upgrade: changeset=2
        migrate: downgrade: changeset=0
        >>> migrate.status()
        Model version:    2
        Database version: 0
        (2, 0)

        """
        if not self.dry_run:
            self.__bootstrap()
            src = self.__database_version()
            dst = self.__model_version()
            self.upgrade(dst)
            self.downgrade(src)

    def status(self):
        """
        Reports the model version and the database version

        >>> # WARNING: SQLite in-memory database seems to not work properly
        >>> #          So, avoid these url: "sqlite://" and "sqlite:///:memory:"
        >>> url = 'sqlite:///build/doctest_status.db'
        >>> import distcontrib.api as api
        >>> migrate = api.migrate(url, 'tests/guinea-pig/admin/migrate')
        >>> migrate.set_verbose(True)
        migrate: application=setup.py repository=tests/guinea-pig/admin/migrate url=sqlite:///build/doctest_status.db
        >>> migrate.status()
        Model version:    2
        Database version: 0
        (2, 0)

        """
        if not self.dry_run:
            self.__bootstrap()
            model_version    = self.__model_version()
            database_version = self.__database_version()
            if self.verbose:
                print('Model version:    %d' % model_version)
                print('Database version: %d' % database_version)
            return ( model_version, database_version )

    def __bootstrap(self):
        import os
        # create repository
        if not os.path.exists(self.repository):
            os.makedirs(os.path.dirname(self.repository))
            dbmigrate.create(self.repository, self.application)
        # put database under version control    
        try:
            self.__database_version()
        except:
            dbmigrate.version_control(self.url, self.repository)

    def __model_version(self):
        return int(dbmigrate.version(self.repository).value)

    def __database_version(self):
        return int(dbmigrate.db_version(self.url, self.repository))


class dsn(object):
    '''
    This class checks, normalizes and store passwords for Data Source Names.
    '''

    def __init__(self, url, dry_run=False, *args, **kwargs):
        '''
        Arguments
        ---------
        url     -- a database connection string, for example:
                       "postgres://username:password@localhost/sample"
                   Note: when the username is not informed, it is retrieved
                         from the environment variable ${USER}
        dry_run -- if True, avoids storing passwords into the standard backing
                   store defined for a given protocol. For example, in the case
                   of Postgres databases, passords are stored in ~/.pgpass
                   ( default: False)
        '''
        super(dsn, self).__init__(*args, **kwargs)
        self.url = url
        self.dry_run = dry_run

    def _invalid(self, what, msg):
        raise RuntimeError('invalid %s: %s' % (what, msg))

    def _simple_database(self, path):
        import re
        d = re.match('^/(\w+)$', path)
        if d is None: self._invalid('database', path)
        return ( '/', d.group(1) )

    def _path_database(self, path):
        dirs=path.split('/')
        return ( '/'.join(dirs[0:-1]) + '/', dirs[-1] )

    def _prompt(self, realm=None, confirm=True):
        import getpass
        ptext = 'Password: ' if realm is None else '[' + realm + '] Password: '
        ctext = 'Confirm: '  if realm is None else '[' + realm + '] Confirm: '
        password = getpass.getpass(ptext)
        confirm  = getpass.getpass(ctext) if confirm else None
        return (password, confirm)

    def _pgsql_check(self, url, protocol, username, hostname, port, database, options):
        if options is not None and len(options) > 0: self._invalid('database', url)

    def _pgsql_passwd(self, hostname, port, database, username, url=None):
        '''
        Allows a password to be defined for a given tuple
        (hostname, port, database, username)

        '''
        import os

        def touch(self, path, mode='a+'):
            fd = os.open(path, os.O_RDWR|os.O_CREAT, 00700)
            return os.fdopen(fd, mode)

        pgpass = os.path.expanduser('~/.pgpass')
        if os.path.exists(pgpass):
            with open(pgpass, 'r') as f:
                for line in f:
                    (h, p, d, u, password) = line.strip().split(':')
                    if not( h == '*' or h  == hostname ): continue
                    if not( p == '*' or int(p) == port ): continue
                    if not( d == '*' or d  == database ): continue
                    if not( u == '*' or u  == username ): continue
                    return password
        # at this point, a password prompt is presented to the user
        ( password, confirm ) = self._prompt(url)
        # if passwords match, write it down to ~/.pgpass
        if password == confirm:
            if not self.dry_run:
                with self.touch(pgpass) as f:
                    f.write('%s:%d:%s:%s:%s' % (hostname, port, database, username, password))
            return password
        else:
            raise RuntimeError('Passwords mismatch')

    _entries = {
                'postgresql': {
                    'hostname': True,
                    'username': True,
                    'password': True,
                    'port'    : 5432,
                    'database': _simple_database,
                    'passwd'  : _pgsql_passwd,
                    'checks'  : _pgsql_check,
                },
                #'sqlite': {
                #    'hostname': False,
                #    'username': False,
                #    'password': False,
                #    'port'    : 0,
                #    'database': _path_database,
                #    'passwd'  : None,
                #    'checks'  : None,
                #},
    }

        
    def check(self):
        '''
        Parses a database url and returns its components.
        Perform protocol specific validations and eventually allows the
        password for the given username to be defined, if allowed by the
        given protocol.

        Arguments
        ---------
        none

        Returns
        -------
        a dictionary containing with components

        Example
        -------
        >>> url = 'postgresql://rgomes:secret@localhost/sample'
        >>> import distcontrib.api as api
        >>> m = api.dsn(url).check()
        >>> m['protocol']
        'postgresql'
        >>> m['username']
        'rgomes'
        >>> m['hostname']
        'localhost'
        >>> m['port']
        5432
        >>> m['database']
        'sample'
        >>> m['options']
        []
        >>> m['password']
        'secret'
        
        '''
        import os
        import getpass
        from urlparse import urlparse

        url = self.url
        u = urlparse(url)
        if u is None: self._invalid('url', url)

        # check protocol, username, hostname and database path
        if u.scheme   is None or len(u.scheme)        < 1: self._invalid('protocol', url)
        if u.scheme in self._entries:
            entry = self._entries[u.scheme]
        else:
            self._invalid('protocol (not implemented yet)', url)

        # check hostname
        if entry['hostname']:
            if u.hostname is None or len(u.hostname)  < 1: self._invalid('hostname', url)
            hostname = u.hostname
        else:
            hostname = ''

        # check database name
        if u.path     is None or len(u.path)          < 2: self._invalid('database', url)

        # url params/query/fragment must be empty
        if u.params   is not None and len(u.params)   > 0: self._invalid('params (must be empty)', url)
        if u.query    is not None and len(u.query)    > 0: self._invalid('query (must be empty)', url)
        if u.fragment is not None and len(u.fragment) > 0: self._invalid('fragment (must be empty)', url)

        # check username
        if entry['username']:
            if u.username is None or len(u.username) < 1:
                username = getpass.getuser()
            else:
                username = u.username
        else:
            username = ''

        # check port number
        if entry['port'] > 0:
            if u.port is not None:
                port = u.port
            else:
                port = entry['port']
        else:
            port = 0

        # find database-specific parameters, if any
        p = u.path.split(';')
        options  = p[1:]
        (path, database) = entry['database'](self, p[0])

        # perform database-specific checkings
        check = entry['checks']
        if check:
            check(self, url, u.scheme, username, u.hostname, port, database, options)

        # retrieve password, if necessary
        if entry['password']:
            if u.password is None or len(u.password) < 1:
                passwd = entry['passwd']
                if passwd is None: self._invalid('protocol (not implemented yet)', url)
                password = passwd(self, u.hostname, port, database, username, url=url)
            else:
                password = u.password
        else:
            password = ''

        # return url components
        d = {}
        d['protocol'] = u.scheme
        d['username'] = username
        d['password'] = password
        d['hostname'] = hostname
        d['port']     = port
        d['path']     = path
        d['database'] = database
        d['options']  = options
        return d

    def normalize(self):
        '''
        Parses a database url and returns a canonical format of it.
        Perform protocol specific validations and eventually allows the
        initial password to be set, if allowed for the given protocol.
        Notice that the password is never returned to the caller.

        Arguments
        ---------
        none

        Returns
        -------
        a normalized URL connection string

        Example
        -------
        >>> url = 'postgresql://localhost/sample'
        >>> import distcontrib.api as api
        >>> api.dsn(url).normalize()
        'postgresql://rgomes@localhost:5432/sample'
        >>> url = 'postgresql://rgomes:secret@localhost/sample'
        >>> api.dsn(url).normalize()
        'postgresql://rgomes@localhost:5432/sample'
        
        '''
        u = self.check()
        opts = u['options']
        if opts is None or len(opts) == 0:
            options = ''
        else:
            options = ';'.join(opts)
        norm   = u['protocol']
        norm  += '://' + u['username']
        #norm += (':' + u['password'])  if len(u['password']) > 0 else ''
        norm  += ('@' + u['hostname'])  if len(u['hostname']) > 0 else ''
        norm  += (':' + str(u['port'])) if u['port'] > 0 else ''
        norm  += u['path']
        norm  += u['database']
        norm  += options
        return norm



def findfile(base='.', names=[], topdown=True, recursive=True, kind=None):
    '''Unix-style find command.

    >>> def printall(items):
    ...     for item in items:
    ...         print(item)
    ...
    >>> import distcontrib.api as api
    >>> printall( api.findfile(kind='f', names=["*.rst", "*.py"], topdown=False) )
    ./tests/guinea-pig/admin/migrate/versions/001_add_initial_tables.py
    ./tests/guinea-pig/admin/migrate/versions/002_rename_eod_columns.py
    ./tests/guinea-pig/admin/migrate/versions/__init__.py
    ./tests/guinea-pig/admin/migrate/pending/implement_options.py
    ./tests/guinea-pig/admin/migrate/manage.py
    ./tests/guinea-pig/admin/migrate/__init__.py
    ./build/lib/distcontrib/api.py
    ./build/lib/distcontrib/command.py
    ./build/lib/distcontrib/__init__.py
    ./distcontrib/api.py
    ./distcontrib/command.py
    ./distcontrib/__init__.py
    ./docs/index.rst
    ./docs/conf.py
    ./tests/conftest.py
    ./tests/test_api.py
    ./CHANGES.rst
    ./README.rst
    ./setup.py
    ./AUTHOR.rst

    '''
    def match(path, names):
        from fnmatch import fnmatchcase
        if names is None or len(names) == 0:
            return True
        fn = os.path.basename(path)
        for name in names:
            if (fnmatchcase(fn, name)):
                return True
        return False

    import os
    import glob
    result = list()
    f = (kind is None or kind=='f')
    d = (kind is None or kind=='d')
    stack = glob.glob(base)
    while stack:
        path = stack.pop(0)
        # print('**** ', path)
        if os.path.isfile(path):
            if f and match(path, names):
                # print('file ', path)
                result.append(path)
        elif recursive and os.path.isdir(path):
            if d and match(path, names):
                # print('dir  ', path)
                result.append(path)
            for item in os.listdir(path):
                stack.append(os.path.join(path,item))
    if not topdown:
        result.reverse()
    return result
